@@ -1,26 +1,24 @@
 package Algorithm::KernelKMeans;
 
 use 5.010;
-use namespace::autoclean;
-use version;
-use Moose;
+use strict;
+use warnings;
 use UNIVERSAL::require;
 
-our $VERSION = '0.03_02';
+our $VERSION = '0.02';
 
-our $IMPLEMENTATION;
-for my $impl (qw/XS PP/) {
-  my $impl_class = __PACKAGE__ . '::' . $impl;
-  if ($impl_class->require) {
-    next if $impl eq 'XS'
-      and version->parse($impl_class->VERSION) < version->parse('0.02_02');
-    $IMPLEMENTATION = $impl_class;
-    extends $IMPLEMENTATION;
-    last;
+BEGIN {
+  for my $impl (qw/XS PP/) {
+    my $impl_class = __PACKAGE__ . '::' . $impl;
+    if ($impl_class->require) {
+      our $IMPLEMENTATION = $impl;
+      parent->use($impl_class); # be child of $impl_class
+      last;
+    }
   }
 }
 
-__PACKAGE__->meta->make_immutable;
+1;
 
 __END__
 
@@ -31,23 +29,20 @@ Algorithm::KernelKMeans - Weighted kernel k-means clusterer
 =head1 SYNOPSIS
 
   use Algorithm::KernelKMeans;
-  use Algorithm::KernelKMeans::Util qw/$KERNEL_POLYNOMINAL/;
-  use List::MoreUtils qw/zip/;
+  use Algorithm::KernelKMeans::Util qw/generate_polynominal_kernel/;
   use Try::Tiny;
   
-  my @vectors = map {
-    my @values = split /\s/;
-    my @keys = 0 .. $#values;
-    +{ zip @keys, @values };
-  } (<>);
+  my @vertices = map { [ split /\s/ ] } (<>);
+  my $kernel = generate_polynominal_kernel(1, 2); # K(x1, x2) = (1 + x1x2)^2
   my $wkkm = Algorithm::KernelKMeans->new( # default weights are 1
-    vectors => \@vectors,
-    kernel => [$KERNEL_POLYNOMINAL => (1, 2)] # K(x1, x2) = (1 + x1x2)^2
+    vertices => \@vertices,
+    kernel => $kernel
   );
   
+  my @clusters;
   try {
-    my $clusters = $wkkm->run(k => 6);
-    for my $cluster (@$clusters) {
+    @clusters = $wkkm->run(k => 6);
+    for my $cluster (@clusters) {
       ...
     }
   } catch {
@@ -67,10 +62,10 @@ This class is just a placeholder. Implementation code is in other class and this
 
 Currently there are 2 implementations: L<Algorithm::KernelKMeans::PP> and L<Algorithm::KernelKMeans::XS>.
 
-C<$Algorithm::KernelKMeans::IMPLEMENTATION> indicates which implementation is loaded.
+C<$Algorithm::KernelKMeans::IMPLEMENTATION> indicates which imlementation is loaded.
 
-Both of these implements same interface (documented below) and C<Algorithm::KernelKMeans> uses faster (XS) implementation if it's available.
-So it's not necessary usually to use the classes directly tough, you can do it if you want.
+Both of these implements same interface (documented below) and C<Algorithm::KernelKMeans> uses faster (XS) imlementation if it's available.
+So it's not neccessary usually to use these classes directly tough, you can do it if you want.
 
 =head1 METHODS
 
@@ -78,49 +73,38 @@ So it's not necessary usually to use the classes directly tough, you can do it i
 
 Constructor. you can specify options below:
 
-=head3 vectors
+=head3 vertices
 
-Required. Array of vectors.
-Each vector is represented as an hash of positive real numbers.
+Required. Array ref of vectors.
+Each vector is represented as an array ref of positive real numbers.
 
 e.g.:
 
  my $wkkm = Algorithm::KernelKMeans->new(
-   vectors => [ +{ prop1 => 229, prop2 => 151, prop3 =>  42 },
-                 +{ prop1 =>  61, prop2 => 151, prop4 => 251 },
-                 +{ prop2 =>  11, prop3 => 120, prop4 =>  55 } ],
-   kernel => [$KERNEL_POLYNOMINAL => (1, 2)]
+   vertices => [ [ 229, 151, 42 ], [ 61, 151, 251 ], [ 11, 120, 55 ] ]
  );
+]
 
 =head3 weights
 
-Array of positive real numbers. Defaults to list of 1s.
+Array ref of positive real numbers. Defaults to list of 1s.
 
 =head3 kernel
 
-Function projects 2 vectors into higher dimentional space and computes inner product.
-
-Kernel function can be specified as a tuple or a code reference.
-
-Tuple is formed with descriptor and parameter(s). For example:
-
-  [$KERNEL_POLYNOMINAL => (1, 2)]
-
-C<$KERNEL_POLYNOMINAL> is a descriptor. And rest of the elements are parameters.
-
-L<Algorithm::KernelKMeans::Util> has some descriptors for some popular kernel functions.
+Kernel function. The function takes 2 vectors and returns 1 positive real number.
+Defaults to K(x1, x2) = (1 + x1x2)^2. L<Algorithm::KernelKMeans::Util> has generators for some popular kernel functions.
 
 =head3 kernel_matrix
 
-2D array of kernel values.
+Array ref of array ref of positive real numbers.
 
 A matrix whose element at (i, j) is K(xi, xj) where i >= j.
 This is derived automatically from C<kernel> by default, however you can specify it manually if you already have it.
 
 Note that the clusterer only uses lower triangle part of the matrix.
-So it is not necessary for the matrix to have element at (i, j) where i < j.
+So it is not neccessary for the matrix to have element at (i, j) where i < j (This argument should be called "kernel triangle" rather than "matrix" probably).
 
-Note that C<kernel> and C<kernel_matrix> are exclusive and either of these is required.
+Note that C<kernel> and C<kernel_matrix> is exclusive. When you specify C<kernel_matrix>, C<kernel> function is never used.
 
 =head2 run(%opts)
 
@@ -133,31 +117,33 @@ Required. (maximum) number of clusters.
 =head3 k_min
 
 Some clusters may be empty during clustering.
-In the case, the clusterer just removes the empty clusters and checks number of rest clusters. If it is less than C<k_min>, the clusterer throws an error.
+In the case of that, the clusterer just removes the empty clusters and checks number of rest clusters. If it is less than C<k_min>, the clusterer throws an error.
 
 Default is same as C<k>.
 
-=head3 initializer
+=head3 shuffle
 
-Specifies cluster initializing method.
-By default, the clusterer initializes clusters using KKZ, which is known as a good initializing procedure.
+When this option is true (this is default), the clusterer sets up initial clusters by random shuffling.
 
-You can C<import> some initializer descriptors from C<Algorithm::KernelKMeans::Util>.
+If it is not what you want, you can set false and get always same result:
+
+  use Test::More;
+  my $clusters1 = $kkm->run(k => 6, shuffle => 0);
+  my $clusters2 = $kkm->run(k => 6, shuffle => 0);
+  my $clusters3 = $kkm->run(k => 6);
+  is_deeply($clusters1, $clusters2); # ok
+  is_deeply($clusters1, $clusters3); # not ok (probably)
 
 =head3 converged
 
 Function predicates that clustering is converged.
 Iteration is broken off and returns result when the predicate returns true.
 
-For each iteration, 2 values will be specified:
+For each iteration, 2 values are specified:
 objective function value of current clusters and new clusters' one.
 As clusters converges, the value decreases.
 
-Default predicate just checks if these 2 values are equal.
-
-=head2 cluster_indices(%opts)
-
-This method is similar to C<run>, but returns clusters contain indices instead of vectors.
+Default predicate just checks if 2 values are equal.
 
 =head1 AUTHOR
 
