@@ -7,39 +7,37 @@ use warnings;
 
 use Carp;
 use List::Util qw/sum shuffle/;
-use List::MoreUtils qw/natatime/;
+use List::MoreUtils qw/natatime each_arrayref/;
 use Moose;
 use MooseX::Types::Common::Numeric qw/PositiveNum/;
-use MooseX::Types::Moose qw/ArrayRef HashRef CodeRef/;
+use MooseX::Types::Moose qw/ArrayRef CodeRef/;
 use POSIX qw/floor/;
 
 use ExtUtils::testlib;
 use Algorithm::KernelKMeans::Util qw/centroid/;
 
-has 'vectors' => (
+has 'vertices' => (
   is => 'ro',
-  isa => ArrayRef[ HashRef[PositiveNum] ],
+  isa => ArrayRef[ ArrayRef[PositiveNum] ],
   required => 1,
   traits => [qw/Array/],
   handles => +{
-    num_vectors => 'count',
-    vector => 'get'
+    num_vertices => 'count',
+    vertex => 'get'
   }
 );
 
 sub sub_vector {
   my ($x1, $x2) = @_;
-  my %tmp; @tmp{keys %$x1, keys %$x2} = ();
-  my %sub = map {
-    my ($e1, $e2) = (($x1->{$_} // 0), ($x2->{$_} // 0));
-    ($_ => $e1 - $e2);
-  } keys %tmp;
-  \%sub;
+  my $iter = each_arrayref $x1, $x2;
+  my @sub;
+  while (my ($s1, $s2) = $iter->()) { push @sub, $s1 - $s2 }
+  \@sub;
 }
 
 sub norm {
   my $vec = shift;
-  sqrt sum map { $_ ** 2 } values %$vec;
+  sqrt sum map { $_ ** 2 } @$vec;
 }
 
 sub score {
@@ -47,16 +45,16 @@ sub score {
   my $score = 0;
   for my $cluster (@$clusters) {
     my $centroid = centroid $cluster;
-    for my $vector (@$cluster) { $score += norm sub_vector($vector, $centroid) }
+    for my $vertex (@$cluster) { $score += norm sub_vector($vertex, $centroid) }
   }
   return $score;
 }
 
 sub init_clusters {
   my ($self, $k, $shuffle) = @_;
-  my @vectors = $shuffle ? shuffle @{ $self->vectors } : @{ $self->vectors };
-  my $n = floor(@vectors / $k);
-  my $iter = natatime $n, @vectors;
+  my @vertices = $shuffle ? shuffle @{ $self->vertices } : @{ $self->vertices };
+  my $n = floor(@vertices / $k);
+  my $iter = natatime $n, @vertices;
   my @clusters;
   while (my @cluster = $iter->()) { push @clusters, \@cluster; }
   if (@{ $clusters[-1] } < $n) {
@@ -70,12 +68,12 @@ sub step {
   my ($self, $clusters) = @_;
   my @centroids = map { centroid $_ } @$clusters;
   my @new_clusters = map { [] } (1 .. @$clusters);
-  for my $vector (map { (@$_) } @$clusters) {
+  for my $vertex (map { (@$_) } @$clusters) {
     my $i = 0;
     my ($nearest) = sort { $a->[1] <=> $b->[1] } map {
-      [ $i++ => norm(sub_vector($vector, $_)) ]
+      [ $i++ => norm(sub_vector($vertex, $_)) ]
     } @centroids;
-    push @{ $new_clusters[$nearest->[0]] }, $vector;
+    push @{ $new_clusters[$nearest->[0]] }, $vertex;
   }
   [ grep { @$_ != 0 } @new_clusters ];
 }
